@@ -15,6 +15,7 @@
 
 #include <bzlib.h>
 #include <zlib.h>
+#include <lzma.h>
 
 #include "util.h"
 #include "md5.h"
@@ -222,13 +223,13 @@ void
 addtocpio(unsigned char **cpiop, bsuint *cpiol, unsigned char *d, int l)
 {
   bsuint cpl = *cpiol;
-  if (cpl + l < cpl || cpl + l + 4095 < cpl + l)
+  if (cpl + l < cpl || cpl + l + 65535 < cpl + l)
     {
       fprintf(stderr, "cpio archive to big\n");
       exit(1);
     }
-  if ((cpl & ~4095) == 0 || (cpl & ~4095) + l > 4096)
-    *cpiop = xrealloc(*cpiop, (cpl + l + 4095) & ~4095);
+  if (cpl == 0 || ((cpl - 1) & 65535) + l >= 65536)
+    *cpiop = xrealloc(*cpiop, (cpl + l + 65535) & ~65535);
   memcpy(*cpiop + cpl, d, l);
   *cpiol = cpl + l;
 }
@@ -347,6 +348,10 @@ str2comp(char *comp)
   if (!strcmp(comp, "gzip rsyncable"))
     return CFILE_COMP_GZ_RSYNC;
 #endif
+  if (!strcmp(comp, "lzma"))
+    return CFILE_COMP_LZMA;
+  if (!strcmp(comp, "xz"))
+    return CFILE_COMP_XZ;
   if (!strcmp(comp, "uncompressed"))
     return CFILE_COMP_UN;
   fprintf(stderr, "unknown compression type: %s\n", comp);
@@ -413,6 +418,18 @@ write_seqfile(struct deltarpm *d, char *seqfile)
     }
 }
 
+static int
+in_multilib_dir(char *fn)
+{
+  char *dirs[] = { "lib/", "lib64/", "lib32/", NULL };
+  int i;
+
+  for (i = 0; dirs[i] != NULL; i++)
+    if (strstr(fn, dirs[i]) != NULL)
+      return 1;
+  return 0;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -422,7 +439,9 @@ main(int argc, char **argv)
   char *nevr;
   int filecnt;
   char **filenames, **filemd5s, **filelinktos;
-  unsigned int *fileflags, *filemodes, *filerdevs, *filesizes, *fileverify;
+  unsigned int *fileflags, *filemodes, *filerdevs, *filesizes, *fileverify, *filecolors;
+  int digestalgo = 1;
+  unsigned int *digestalgoarray;
   int i, fd, l, l2, l3;
   struct cfile *bfd;
   struct cpiophys cph;
@@ -439,6 +458,7 @@ main(int argc, char **argv)
   int skipped_badsize = 0;
   int skipped_fileflags = 0;
   int skipped_verifyflags = 0;
+  int skipped_multilib = 0;
   int skipped_all = 0;
   int pinfo = 0;
   struct rpmlfile *files1 = 0;
@@ -483,6 +503,7 @@ main(int argc, char **argv)
   int paycomp = CFILE_COMP_XX;
   int addblkcomp = CFILE_COMP_BZ;
   int targetcomp = CFILE_COMP_XX;
+  char *payloadflags;
 
   memset(&d, 0, sizeof(d));
   while ((c = getopt(argc, argv, "vV:prl:s:z:u")) != -1)
@@ -562,7 +583,7 @@ main(int argc, char **argv)
       nevr1 = rpmlread(pfp, argv[argc - 5 + alone], 0, &files1, &nfiles1);
       fclose(pfp);
       pfd = open(argv[argc - 4 + alone], O_RDONLY);
-      if (pfp < 0)
+      if (pfd < 0)
 	{
 	  perror(argv[argc - 4 + alone]);
 	  exit(1);
@@ -685,6 +706,8 @@ main(int argc, char **argv)
       targetcomp = CFILE_COMP_UN;
       if (paycomp == CFILE_COMP_XX)
 	paycomp = CFILE_COMP_GZ;	/* no need for better compression */
+      if (addblkcomp == CFILE_COMP_XX)
+	addblkcomp = CFILE_COMP_GZ;
       d.targetsize = fullsize;
       d.targetcomp = targetcomp;
       d.targetcomppara = 0;
@@ -735,6 +758,18 @@ main(int argc, char **argv)
   filemodes = headint16(h, TAG_FILEMODES, (int *)0);
   fileverify = headint32(h, TAG_FILEVERIFY, (int *)0);
   filelinktos = headstringarray(h, TAG_FILELINKTOS, (int *)0);
+  filecolors = headint32(h, TAG_FILECOLORS, (int *)0);
+
+  if ((digestalgoarray = headint32(h, TAG_FILEDIGESTALGO, (int *)0)))
+    {
+      digestalgo = digestalgoarray[0];
+      free(digestalgoarray);
+    }
+  if (digestalgo != 1 && digestalgo != 8)
+    {
+      fprintf(stderr, "Unknown digest type: %d\n", digestalgo);
+      exit(1);
+    }
 
   if (alone)
     {
@@ -760,8 +795,13 @@ main(int argc, char **argv)
 	  exit(1);
 	}
       targetcomp = newbz->comp;
+      if ((payloadflags = headstring(d.h, TAG_PAYLOADFLAGS)) != 0)
+	if (*payloadflags >= '1' && *payloadflags <= '9')
+	  targetcomp = cfile_setlevel(targetcomp, *payloadflags - '0');
       if (paycomp == CFILE_COMP_XX)
 	paycomp = targetcomp;
+      if (addblkcomp == CFILE_COMP_XX)
+	addblkcomp = targetcomp;
       while ((l = newbz->read(newbz, buf, sizeof(buf))) > 0)
 	addtocpio(&newcpio, &newcpiolen, (unsigned char *)buf, l);
       if (l < 0)
@@ -802,7 +842,6 @@ main(int argc, char **argv)
     fprintf(vfp, "reading old rpm...\n");
   if (rpmonly)
     {
-      write(3, oldcpio, oldcpiolen);
       while ((l = bfd->read(bfd, buf, sizeof(buf))) > 0)
 	addtocpio(&oldcpio, &oldcpiolen, (unsigned char *)buf, l);
     }
@@ -893,6 +932,12 @@ main(int argc, char **argv)
 		    fprintf(vfp, "skipping %s: bad verify flags %x\n", np, fileverify[i]);
 		  skipped_verifyflags++;
 		}
+	      else if (filecolors && (filecolors[i] & (RPMFC_ELF32 | RPMFC_ELF64)) != 0 && !in_multilib_dir(np))
+		{
+		  if (verbose > 1)
+		    fprintf(vfp, "skipping %s: colored file in non-multilib dir\n", np);
+		  skipped_multilib++;
+		}
 	      else
 		{
 		  if (verbose > 1)
@@ -995,9 +1040,17 @@ oaretry1:
 		}
 	      if (S_ISREG(filemodes[i]) && lsize)
 		{
-		  unsigned char fmd5[16];
-		  parsemd5(filemd5s[i], fmd5);
-		  rpmMD5Update(&seqmd5, fmd5, 16);
+		  unsigned char fmd5[32];
+		  if (digestalgo == 1)
+		    {
+		      parsemd5(filemd5s[i], fmd5);
+		      rpmMD5Update(&seqmd5, fmd5, 16);
+		    }
+		  else
+		    {
+		      parsesha256(filemd5s[i], fmd5);
+		      rpmMD5Update(&seqmd5, fmd5, 32);
+		    }
 		}
 	      addtoseq(i);
 	    }
@@ -1049,6 +1102,8 @@ oaretry1:
 	    fprintf(vfp, "  file flags:   %4d/%d = %.1f%%\n", skipped_fileflags, skipped_all, skipped_fileflags * 100. / skipped_all);
 	  if (skipped_verifyflags)
 	    fprintf(vfp, "  verify flags: %4d/%d = %.1f%%\n", skipped_verifyflags, skipped_all, skipped_verifyflags * 100. / skipped_all);
+	  if (skipped_multilib)
+	    fprintf(vfp, "  colored-not-in-multidir: %4d/%d = %.1f%%\n", skipped_multilib, skipped_all, skipped_multilib * 100. / skipped_all);
 	}
       addtoseq(-1);
       if (verbose > 1)
@@ -1073,6 +1128,7 @@ oaretry1:
   fileverify = xfree(fileverify);
   filelinktos = xfree(filelinktos);
   filenames = xfree(filenames);
+  filecolors = xfree(filecolors);
   if (!alone)
     h = xfree(h);
 
@@ -1138,8 +1194,13 @@ oaretry1:
 	  exit(1);
 	}
       targetcomp = newbz->comp;
+      if ((payloadflags = headstring(d.h, TAG_PAYLOADFLAGS)) != 0)
+	if (*payloadflags >= '1' && *payloadflags <= '9')
+	  targetcomp = cfile_setlevel(targetcomp, *payloadflags - '0');
       if (paycomp == CFILE_COMP_XX)
 	paycomp = targetcomp;
+      if (addblkcomp == CFILE_COMP_XX)
+	addblkcomp = targetcomp;
       while ((l = newbz->read(newbz, buf, sizeof(buf))) > 0)
 	addtocpio(&newcpio, &newcpiolen, (unsigned char *)buf, l);
       if (l < 0)
